import os, sys, math
from timeit import default_timer as timer

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

# calibration object takes care of storing calibration coordinates and
# drawing the calibration animation 
#
# calibration is performed by constricting and dilating a small dot
# at each of several calibration coordinates (when the eye tracker samples
# the camera to get the eye-in-frame coordinates)
#
# following sampling, the calibration dot is animated from location to
# location
#
# CALIBRATE_DELAY: time between stationary calibration dots
# CALIBRATE_TIMEOUT: time to dilate and constrict dot (the sampling time)
#
class Calibration:
  
  def __init__(self):
    self.dot_index = 0
    self.dot_radius = 5.0
    self.movement = 0
    self.moveTime = 1000.0 # CALIBRATE_DELAY
    self.zoomTime = 750.0 # CALIBRATE_TIMEOUT / 2 (dilate and constrict)
    self.innerR = 5.0
    self.outerR = 20.0
    self.elapsed = 0.0
    self.vertices = []
    self.xy = (0,0)
    self.st = 0.0
    self.et = 0.0
    self.tt = 0.0

    self.append_vertices()

  def timer_start(self):
    self.elsapsed = 0.0
    self.st = timer()

  def timer_end(self):
    self.et = timer()
    self.tt = self.et - self.st

  def timer_elapsed(self):
    return timer() - self.st

  def timer_elapsed_ms(self):
    return (timer() - self.st) * 1000.0

  def append_vertices(self):

    # point 0, origin
    self.vertices.append((0,0))

    # specify calibration points in tracker's ref frame with (0,0) at top left
    cx = .5
    cy = .5

    # point 1, UTL
    x = cx - .33
    y = cy - .33
    self.vertices.append((x,y))

    # point 2, UTM
#   x = cx
#   y = cy - .33
#   self.vertices.append((x,y))

    # point 3, UTR
    x = cx + .33
    y = cy - .33
    self.vertices.append((x,y))

    # point 4, LTR
#   x = cx + .33
#   y = cy
#   self.vertices.append((x,y))

    # point 5, LTM
    x = cx
    y = cy
    self.vertices.append((x,y))

    # point 6, LTL
#   x = cx - .33
#   y = cy
#   self.vertices.append((x,y))

    # point 7, LLL
    x = cx - .33
    y = cy + .33
    self.vertices.append((x,y))

    # point 8, LLM
#   x = cx
#   y = cy + .33
#   self.vertices.append((x,y))

    # point 9, LLR
    x = cx + .33
    y = cy + .33
    self.vertices.append((x,y))

  def get_calibration_points(self):
    return self.vertices

  def get_xy(self):
    return self.xy

  def set_xy(self, xxx_todo_changeme):
    (x,y) = xxx_todo_changeme
    self.xy = (x,y)

  def get_dot_len(self):
    return len(self.vertices)

  def get_dot_index(self):
    return self.dot_index

  def inc_dot_index(self):
    self.dot_index += 1
    if self.dot_index >= len(self.vertices):
      self.dot_index = 0

  def get_dot(self,i):
    return self.vertices[i]

  def get_dot_coord_x(self,i):
    return self.vertices[i][0]
    
  def get_dot_coord_y(self,i):
    return self.vertices[i][1]

  def get_dot_radius(self):
    return self.dot_radius

  def set_dot_radius(self,r):
    self.dot_radius = r

  def get_zoomTime(self):
    return self.zoomTime

  def get_moveTime(self):
    return self.moveTime

  def get_innerR(self):
    return self.innerR

  def get_outerR(self):
    return self.outerR
