# Author Darrell Best: bbest@g.clemson.edu
# Date 10/06/2014
#
# Singleton State class holds global variables

import sys, math
import time
from collections import deque

from rstats import RunningStats

# prevent Python from writing .pyc files
sys.dont_write_bytecode = True

class State:
  
    def __init__(self):
        self.mouse_coord = [.5, .5]
        self.gaze_coord = [.5, .5]
        self.fullscreen = 0
        self.butterworth = 0
        self.input = 0 # 1 is eye tracker, 0 is mouse
        self.dfwidth = 10
        self.dfdegree = 2
        self.dfo = 1
        self.T1 = 3
        self.T2 = 10
        self.event = 0		# 0: saccade  1: fixate  2: pursuit
        self.st = 0.0		# current fixation start time
        self.et = 0.0		# current fixation end time
        self.tt = 0.0		# current fixation total time (duration)
        self.ux = 0.0		# current fixation mean x coord
        self.uy = 0.0		# current fixation mean y coord
        self.sx = 0.0		# current fixation x coord stdev
        self.sy = 0.0		# current fixation y coord stdev
        self.k = 0.0		# current fixation count
        self.uamp = 0.0		# current mean inter-fixation amplitude
        self.samp = 0.0		# current mean -fixation stdev
        self.udur = 0.0		# current mean fixation duration
        self.sdur = 0.0		# current mean fixation stdev
        self.K = 0.0		# current K coefficient
        self.rx = RunningStats()# current running statistics for x coord
        self.ry = RunningStats()# current running statistics for y coord
        self.points = deque(maxlen=30)
        self.fixations = deque(maxlen=30)
        self.vms = 0.0
        self.vtd = 0.0

    def set_input(self, flag):
        self.input = flag

    def set_mouse_coord_x(self, x):
        self.mouse_coord[0] = x

    def set_mouse_coord_y(self, y):
        self.mouse_coord[1] = y

    def set_gaze_coord_x(self, x):
        self.gaze_coord[0] = x

    def set_gaze_coord_y(self, y):
        self.gaze_coord[1] = y

    def toggle_butterworth(self):
        self.butterworth = 1 - self.butterworth

    def set_butterworth(self, flag):
        self.butterworth = flag

    def set_dfwidth(self,d):
        self.dfwidth = d

    def set_dfdegree(self,d):
        self.dfdegree = d

    def set_dfo(self,d):
        self.dfo = d

    def set_T1(self,T):
        self.T1 = T

    def set_T2(self,T):
        self.T2 = T

    def set_event(self,e):
        self.event = e

    def get_input(self):
        return self.input

    def get_mouse_coord_x(self):
        return self.mouse_coord[0]
    
    def get_mouse_coord_y(self):
        return self.mouse_coord[1]
    
    def get_gaze_coord_x(self):
        return self.gaze_coord[0]
    
    def get_gaze_coord_y(self):
        return self.gaze_coord[1]
    
    def get_butterworth(self):
        return self.butterworth
    
    def get_dfwidth(self):
        return self.dfwidth

    def get_dfdegree(self):
        return self.dfdegree

    def get_dfo(self):
        return self.dfo

    def get_T1(self):
        return self.T1

    def get_T2(self):
        return self.T2

    def get_event(self):
        return self.event

    def set_fullscreen(self,flag):
        self.fullscreen = flag

    def get_fullscreen(self):
        return self.fullscreen

    def toggle_fullscreen(self):
        self.fullscreen = 1 - self.fullscreen

    def end_fixation(self,monitor,et):
      # get end time of current point, compute duration
      self.et = et
      self.tt = self.et - self.st
      # don't add fixation with -ve duration
      # (due to spurious timestamp in .raw file)
      if self.st > 0.0 and self.tt > 0.0:
        # need to clamp fixation centroid: it could have gone negative
        # only do this if we're dealing with normalized coords, which we're not
        # (maybe we should be)
#       self.ux, self.uy = clamp(self.ux, self.uy)

        # if not first fixation, check amp between this and previous fixation
        if len(self.fixations):
          px, py = self.fixations[-1].get()
          amp = math.sqrt((self.ux-px)**2 + (self.uy-py)**2)
          amp = monitor.pix2deg(amp)
          dur = self.tt

          l = len(self.fixations)
          self.udur = float(l)/float(l+1) * self.udur + \
                           1.0/float(l+1) * dur
          self.sdur = math.sqrt(float(l-1)/float(l) * self.sdur**2 + \
                                       1.0/float(l+1) * (dur-self.udur)**2)
          self.uamp = float(l)/float(l+1) * self.uamp + \
                           1.0/float(l+1) * amp
          self.samp = math.sqrt(float(l-1)/float(l) * self.samp**2 + \
                                       1.0/float(l+1) * (amp-self.uamp)**2)
          self.K = (dur - self.udur)/self.sdur - \
                   (amp - self.uamp)/self.samp
#         print("ux, uy: %5.2f %5.2f" % (self.ux, self.uy))
#         print("x, y: ", x, y)
#         print("px, py: ", px, py)
#         print("amp: ", amp)
#         print("uamp: ", self.uamp)
#         print("samp: ", self.samp)
#         print("udur: ", self.udur)
#         print("sdur: ", self.sdur)
#         print("K: ", self.K)

        # add current fixation to deque
        self.fixations.append(Fixation(self.ux,self.uy,self.st,self.tt))

      # saccade starts, reset fixation timers
      self.et = 0.
      self.tt = 0.

    def start_fixation(self,st):
      # set start time
      self.st = st

      # reset running mean and counter
      self.ux = 0.
      self.uy = 0.
      self.sx = 0.
      self.sy = 0.
      self.k = 0

    def cont_fixation(self,x,y):
      # compute running mean
      # from Brown, Robert Grover, "Introduction to Random Signal
      #   Analysis and Kalman Filtering", John Wiley & Sons, New York, NY
      #   1983 [p.182] TK5102.5 .B696
  
      self.ux = float(self.k)/float(self.k+1) * self.ux + \
                          1.0/float(self.k+1) * x
      self.uy = float(self.k)/float(self.k+1) * self.uy + \
                          1.0/float(self.k+1) * y

      # see: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
      # and look at Welford's online algorithm
      if self.k:
        # this would be computing s^2 the sample variance
        self.sx = math.sqrt(float(self.k-1)/float(self.k) * self.sx**2 + \
                                        1.0/float(self.k+1) * (x-self.ux)**2)
        self.sy = math.sqrt(float(self.k-1)/float(self.k) * self.sy**2 + \
                                        1.0/float(self.k+1) * (y-self.uy)**2)

        # this would be computing sigma^2 the population variance
        # need ux, uy from above
#       self.sx = math.sqrt(float(self.k)/float(self.k+1) * self.sx**2 + \
#                                     1.0/float(self.k+1) * (x-self.ux)*(x-ux))
#       self.sy = math.sqrt(float(self.k)/float(self.k+1) * self.sy**2 + \
#                                     1.0/float(self.k+1) * (y-self.uy)*(y-uy))

      self.k += 1
